'''
Created on 10 Sep 2009

@author: Paul Chapman
'''

import re

class _Writer(object):

    def _addrun(self, n, c):
         self.rle += (str(n) if n != 1 else "") + c
         
    def _addlives(self):
        if self._lives > 0:
            self._addrun(self._lives, "o")
            self._lives = 0
        
    def __init__(self, cells):

        self.rle = ""
        self._lives = 0
        
        points = [(cells[i + 1], cells[i]) for i in xrange(0, len(cells), 2)]
        points.sort()

        xc, yc = 0, 0
        
        for y, x in points:
            if y > yc:
                self._addlives()
                self._addrun(y - yc, "$")
                xc, yc = 0, y
            if x > xc:
                self._addlives()
                self._addrun(x - xc, "b")
                xc = x
            self._lives += 1
            xc += 1
            
        self._addlives()
        self.rle += "!"


class _Reader(object):

    _REP = re.compile(r"""
        ( \d*[bo\$]  # old-style run
        | \@\w+      # name
        | \+\d+      # evolve
        | . )        # delimeter
        """, re.VERBOSE)
    
    _ID_CHAR = "."
    _XFORM_CHARS = _ID_CHAR + r"<^>/-\|"
    _TERMINATOR_CHARS = _XFORM_CHARS + "+"

    _MATRICES = {
        _ID_CHAR: ( 1,  0,  0,  1),
             ">": ( 0,  1, -1,  0),
             "^": (-1,  0,  0, -1),
             "<": ( 0, -1,  1,  0),
             "|": (-1,  0,  0,  1),
             "-": ( 1,  0,  0, -1),
             "/": ( 0, -1, -1,  0),
            "\\": ( 0,  1,  1,  0)
        }

    def __init__(self, string, evolve, caching):
        self._evolve = evolve
        self._caching = caching
        self._cache = {}
        string = re.sub(r"\s+", "", string)
        self._iterator = self._REP.finditer(string)
        token, self.cells = self._parse(self._next(), "!")

    def _next(self):
        try: return self._iterator.next().group()
        except StopIteration: return None

    def _parse(self, token, terminators):
        cells = []
        x = y = 0
        while token != None:
            if token.startswith("@"):
                subpattern = self._subpattern(token[1:])
                cells.extend(self._xlate(subpattern, x, y))
            elif token[-1:] in "bo$":
                n = int(token[:-1]) if len(token) > 1 else 1
                if token[-1:] == "$":
                    y += n
                    x = 0
                else:
                    if token[-1:] == "o":
                        for i in xrange(n):
                            cells.extend([x + i, y])
                    x += n
            else:
                if token not in terminators:
                    raise SyntaxError, "Syntax error in rle"
                return token, cells
            token = self._next()
        raise SyntaxError, "Syntax error in rle"
            
    def _subpattern(self, name):
        token = self._next()
        namekey = name + "+0" + self._ID_CHAR
        if token == "=":
            token, self._cache[namekey] = self._parse(self._next(),
                                                      self._TERMINATOR_CHARS)
        if token.startswith("+") and len(token) > 1:
            n = int(token[1:])
            token = self._next()
        else:
            n = 0
        if token not in self._XFORM_CHARS:
            raise SyntaxError, "Syntax error in rle"
        resultkey = name + "+" + str(n) + token
        try:
            return self._cache[resultkey]
        except KeyError:
            pass
        basekey = name + "+0" + token
        try:
            base = self._cache[basekey]
        except KeyError:
            base = self._xform(self._cache[namekey], token)
            if self._caching != 0:
                self._cache[basekey] = base
        result = self._gen(base, n)
        if n != 0 and self._caching == 2:
            self._cache[resultkey] = result
        return result
    
    def _xlate(self, cells, xt, yt):
        if xt == yt == 0:
            return cells
        result = list(cells)
        for i in xrange(0, len(result), 2):
            x, y = result[i:i+2]
            result[i:i+2] = x + xt, y + yt
        return result
    
    def _xform(self, cells, char):
        if char == self._ID_CHAR:
            return cells
        matrix = self._MATRICES[char]
        result = list(cells)
        xm = ym = 0
        for i in xrange(0, len(result), 2):
            x, y = result[i:i+2]
            xn, yn = (matrix[0] * x + matrix[1] * y,
                      matrix[2] * x + matrix[3] * y)
            xm, ym = min(xm, xn), min(ym, yn)
            result[i:i+2] = xn, yn
        return self._xlate(result, max(0, -xm), max(0, -ym))
    
    def _gen(self, cells, n):
        if n == 0:
            return cells
        if self._evolve is None:
            raise SyntaxError, "'+' used without supplied evolve function"
        return self._evolve(cells, n)
    
    
def rle(cells):
    return _Writer(cells).rle

def cells(str, evolve=None, caching=0):
    return _Reader(str, evolve, caching).cells

__all__ = ["rle", "cells"]

if __name__ == "__main__":
    from Try import Pattern
    def evolve(cells, n):
        return Pattern(cells).evolved(n).cells
    print cells("@A=3o$o$bo.10b@A+1^!", evolve)
