import cProfile
import random
# import build

from gemini import Assessor, Salvo, G, ResultDict
from gemini import isNone, isNotNone, isAny, isZero, isNegative, isPositive
from rlife import X, BLOCK, BEEHIVE, BLINKER, BOAT
from gemini import Foo, TimeStr, Operation

import results
import string

random.seed()

def LanesToSalvos(lanes):
    return tuple(map(lambda x: Salvo(G(x, 0)), lanes))
    
def Assess(elbow, lanes, essentialMinutes, bonusMinutes):
    assessor = Assessor(elbow, LanesToSalvos(lanes),
                        results.Searches,
                        essentialMinutes, bonusMinutes)
    return assessor.run()

def RandomLaneSet(m, n):
    lanes = sorted(random.sample(range(m, n), 3))
    return lanes

def RandomLaneSet2():
    while True:
        hitlane = random.choice([-5, -4, -3, 2, 3, 4])
        mylane = random.choice([-9, -8, -7, 6, 7, 8])
        #mylane = random.choice([-12, 11])
        inlane = random.choice(range(-8, 8))
        lanes = sorted([inlane, mylane, hitlane])
        if len(set(lanes)) == 3: break
    return lanes

def RandomLaneSet3():
    while True:
        hitlane = random.choice([-5, -4, -3, 2, 3, 4])
        mylanes = random.sample(range(-6, 6), 2)
        lanes = sorted([hitlane] + mylanes)
        if len(set(lanes)) == 3: break
    return lanes

def RandomLaneSet4():
    while True:
        hitlane = random.choice([-5, -4, -3, 2, 3, 4])
        lane = random.choice(range(-8, 8))
        duplane = random.choice([hitlane, lane])
        lanes = sorted([hitlane, lane, duplane])
        if len(set(lanes)) == 2: break
    return lanes

def RandomBeehiveLaneSet1():
    while True:
        hitlane = random.choice([-5, -4, -3, -2, -1, 0, 2, 3, 6])
        lanes = sorted(random.sample(range(-5, 5), 2) + [hitlane])
        if len(set(lanes)) == 3: break
    return lanes

def RandomBlinkerLaneSet1():
    lanes = sorted(random.sample(range(-5, 9), 3))
    return lanes

def NewPrintSummary():
    result = []
    for k, v in results.Searches.iteritems():
        if v:
            ops = ResultDict(list)
            ops.update(v["operations"])
            keys = ops.keysMatching(isNone, isNegative, isZero)
            decs = sorted((abs(k.inc) for k in keys), reverse=True)
            keys = ops.keysMatching(isNone, isPositive, isZero)
            incs = sorted(k.inc for k in keys)
            keys = ops.keysMatching(isNotNone, isAny, isZero)
            #outputs = sorted(k.lane for k in keys)
            #if keys:
            keys = sorted(keys, key=lambda key: key.lane)

#                cycles = [len(ops[key][0]._contents) for key in keys]
#                outputs = [str(key.lane) +
#                           "*" * (len(ops[key][0]._contents) - 1)
#                           for key in keys]
#                outputs = " ".join(outputs)

            pairs = [[len(ops[key][0]._contents), key.lane]
                       for key in keys]
            pairs1c = filter(lambda pair: pair[0] == 1, pairs)
            pairs2c = filter(lambda pair: pair[0] == 2, pairs)
            out1c = [x[1] for x in pairs1c]
            out2c = [x[1] for x in pairs2c]
            
            lanes = [s._contents[0].x for s in k[1]]
            if decs and incs:
                result += [[lanes, decs, incs, out1c, out2c]]
    f = lambda x, y: len(x[1]) - len(y[1])
    result = sorted(result, f, reverse=True)
    g = lambda x: " ".join(str(y) for y in x)
    fmt = "{0:>12}{1:>12}   {2:12}{3:>8}   {4:8}"
    print "           All Results with One-Cycle INCs and DECs"
    print
    print fmt.format("Salvos", "DECs", "INCs", "1C Out", "2C Out", )
    print fmt.format("------", "----------", "----------", "------", "------")
#    for lanes, decs, incs, outputs in result:
#        print fmt.format(g(lanes), g(decs), g(incs), outputs)
    for l in result:
        print fmt.format(*map(g, l))

def PrintSummary():
    
    def f(key, n):
        #s = string.ascii_uppercase[len(ops[key][0]._contents) - 1].replace("C", "")
        s = key.quickStr(ops[key][0]).replace("C", "")
        return str(n) + "".join(s)
    
    result = []
    for k, v in results.Searches.iteritems():
        if v:
            ops = ResultDict(list)
            ops.update(v["instructions"])
            if ops.isComplete():
                keys = ops.keysMatching(isNone, isNone, isNegative)
                keys = sorted(keys, key=lambda op: op.inc)
                decs = [f(op, -op.inc) for op in keys]
                keys = ops.keysMatching(isNone, isNone, isPositive)
                keys = sorted(keys, key=lambda op: op.inc)
                incs = [f(op, op.inc) for op in keys]
                keys = ops.keysMatching(isNotNone, isPositive, isAny)
                keys = sorted(keys, key=lambda op: op.lane)
                outputs = [f(op, -op.lane) for op in keys]
                oincs = [op.inc for op in keys]
                oo = zip(outputs, oincs)
                oo = [o[0] + "(" + str(o[1])+ ")" for o in oo]
                lanes = [s._contents[0].x for s in k[1][:2]]
                result += [[lanes, decs, incs, oo]]
    f = lambda x, y: len(x[1]) - len(y[1])
    result = sorted(result, f, reverse=True)
    g = lambda x: " ".join(str(y) for y in x)
    fmt = "{0:>6} {1:>36}  {2:12}{3:12}"
    print fmt.format("Salvos", "DECs", "INCs", "Outputs")
    print fmt.format("------", "----", "----", "-------")
    for lanes, decs, incs, outputs in result:
        print fmt.format(g(lanes), g(decs), g(incs), g(outputs))

def FormatOps(ops):
    keys = sorted(ops.keysMatching(isNone, isNone, isNegative))
    decs = [k.quickStr(ops[k][0]) for k in keys]
    keys = sorted(ops.keysMatching(isNone, isNone, isPositive))
    incs = [k.quickStr(ops[k][0]) for k in keys]
    keys = sorted(ops.keysMatching(isNotNone, isZero, isAny))
    loutputs = [k.quickStr(ops[k][0]) for k in keys]
    keys = sorted(ops.keysMatching(isNotNone, isPositive, isAny))
    routputs = [k.quickStr(ops[k][0]) for k in keys]
    return decs, incs, loutputs, routputs

def PrintConstructibles(*lanes):
    g = lambda x: " ".join(str(y) for y in x)
    print
    print TimeStr(), "Constructible salvos for lanes", lanes[:2], "..."
    constructibles = Foo(lanes, results.Searches)
    if constructibles:
        constructibles = sorted(constructibles,
                                key=lambda x: FormatOps(x[2]))
        print " Inserters  Order    Operations"
        for names, permlanes, ops in constructibles:
            print " {0:9}  {1:9}".format("-".join(names),
                                         tuple(permlanes)),
            for s, z in zip(("  DECs", "  INCs", " LEFTs", "RIGHTs"),
                         FormatOps(ops)):
                print s + ": " + g(z)
                print "                     ",
            print
            #print " {0}  {1}  {2}  {3}".format(*map(g, FormatOps(ops)))
    print TimeStr(), "Done"

def Random2LaneBlinkerSet1():
    while True:
        hitlane = random.choice(range(-5, 7))
        lane = random.choice(range(-20, 21))
        lanes = sorted([hitlane, lane]) + [100]
        if len(set(lanes)) == 3: break
    return lanes

def Random2LaneBeehiveSet1():
    while True:
        hitlane = random.choice(range(-6, 7))
        lane = random.choice(range(-20, 21))
        lanes = sorted([hitlane, lane]) + [100]
        if len(set(lanes)) == 3: break
    return lanes

def Random2LaneBoatSet1():
    while True:
        hitlane = random.choice(range(-6, 6))
        lane = random.choice(range(-20, 20))
        lanes = sorted([hitlane, lane]) + [100]
        if len(set(lanes)) == 3: break
    return lanes

#PrintSummary()

def FixMisfiledResults():
    for k, v in results.Searches.iteritems():
        if v:
            assessor = Assessor(k[0], k[1], results.Searches, 0, 0)
            assessor.fixMisfiledResults()
            reload(results)

def Random2LaneBlockSet1():
    while True:
        hitlane = random.choice(range(-6, 6))
        lane = random.choice(range(-10, 10))
        lanes = sorted([hitlane, lane])
        lanes2 = sorted(-1 - lane for lane in lanes)
        if lanes2 < lanes:
            lanes = lanes2
        lanes += [100]
        if len(set(lanes)) == 3: break
    return lanes

def Random2LaneBlockSet2():
    while True:
        hitlane = random.choice(range(-6, 6))
        lane = random.choice(range(-8, 8))
        lanes = sorted([hitlane, lane])
        if len(set(lanes)) == 1: continue
        lanes2 = sorted(-1 - lane for lane in lanes)
        if lanes2 < lanes: continue
        if lanes == [-5, -2] or lanes == [-5, 2]: continue
        lanes += [100]
        break
    return lanes

def Goo(elbow, *lanes):
    r = Assess(elbow, lanes, 10000, 10000)
    reload(results)
    if r and r.isComplete():
        PrintConstructibles(*lanes)
    print ("####################################"
           "####################################")

Goo(BLOCK, -5, 5, 100)

# while True:
#     elbow = BLOCK
#     lanes = Random2LaneBlockSet2()
#     if results.Searches.get((elbow, LanesToSalvos(lanes))) is None:
#         Goo(elbow, *lanes)


#Goo(BLOCK, -10, -3, 100)

# -5 0 (-1 4)
# -6 4 (-5 5)
# -3 -2 (1 2)
# -6 -4 (3 5)
# -5 2
# -10 -3

#Goo(BLOCK, -5, 02, 100)
#Goo(BLOCK, -5, 2, 100)
#PrintSummary()
#PrintConstructibles(-5, -2, 100)
#PrintConstructibles(-5, 2, 100)


#Goo(BLOCK, -8, -5, 100)

#cProfile.run("PrintConstructibles(-5, -3, 2)", sort=1, filename="cons532B.cprof")
        
#Goo(BLOCK, -8, -5, 5)

#Goo(BLOCK, -5, -3, 2)

#Goo(-3, -2, 2)
#Goo(2, 4, 5)

#PrintConstructibles(-5, 0, 100)
#PrintConstructibles(-6, 4, 100)
#PrintConstructibles(-3, -2, 100)
#PrintConstructibles(-6, -4, 100)
#PrintConstructibles(-5, 2, 100)
#PrintConstructibles(-10, -3, 100)

#Goo(-3, -2, 2)
#Goo(-5, -3, 2)
#Goo(-7, -4, -2)
#Goo(-3, 1, 4)
#Goo(-6, -4, 7)
#Goo(-2, 3, 4)
#Goo(-4, 4, 6)
#Goo(2, 4, 9)
#Goo(-5, -4, 5)
#Goo(-5, -4, 1)
#Goo(-4, -1, 4)
#Goo(-5, 4, 7)
#Goo(-6, 0, 4)

#Foo((-3, -2, 2), results.Searches) # -3 1 / -11 -2
#Foo((-6, -3, 4), results.Searches) # -5 -3 -2 **1 / **-9 **4
#Foo((-5, -3, 4), results.Searches) # -3 1 **2 / **0 **-13
#Foo((-10, -5, 5), results.Searches) # **-8 **-4 -3 **-2 1 / **-3 **4
#Foo((-4, 4, 12), results.Searches) # **-7 -4 1 / **-7 **-6

#PrintConstructibles(Foo((-7, -4, -2), results.Searches))
#PrintConstructibles(Foo((-5, 4, 5), results.Searches))
#PrintConstructibles(Foo((-5, -3, -2), results.Searches))

#PrintConstructibles(Foo((-3, -2, 2), results.Searches))
#PrintConstructibles(Foo((-6, -3, 4), results.Searches))
#PrintConstructibles(Foo((-5, -3, 4), results.Searches))


#PrintConstructibles(2, 4, 5)

# -3 1 / -11(-9) -2(-5)
# -5 -3 -2 1B / -9(1)B 4(-11)B
# -3 1 2B / -13(1)B 0(-9)B
# -8B -4B -3 -2B 1 / -3(3)B 4(2)B
# -7B -4 1 / -7(1)B -6(0)B

#Foo((2, 4, 5), results.Searches) # -7,-4,2/-14,-17

#Assess(BLOCK, (Salvo(G(2, 0)), Salvo(G(4, 0)), Salvo(G(5, 0))), 10000)
#Assess(BLOCK, tuple(map(lambda x: Salvo(G(x, 0)), (2, 4, 5))), 10000)
#Assess(BLOCK, tuple(map(lambda x: Salvo(G(x, 0)), (2, 4, 9))), 10000)
#Assess(BLOCK, tuple(map(lambda x: Salvo(G(x, 0)), (-3, -2, 2))), 10000)

#try:
#    import golly as g
#    a = [GGSeq[n] for n in range(100)]
#    p = StampCollection(a)
#    g.addlayer()
#    g.setname("Stamp collection of " + str(len(a)))
#    g.setrule("B3/S23")
#    g.putcells(list(p.cells))
#    g.fit()
#except ImportError:
#    a = [GGSeq[n] for n in range(5)]
#    print [p._pst for p in a]
#    inc = GX2(-5, 30)
#    black = GX2(-13, 34) + GX2(-10, 60)
#    white = GX2(-16, 19) + GX2(-15, 35)
#    newwhite = GX2(-15, 29) + GX2(-19, 65)
#    print dec
#    print inc
#    print MinFollowDelays(dec, dec + inc)
#    for i in range(20, 30):
#        n = 1 << i
#        salvo = GGSeq[n]
#        sp = sum(salvo.dbox[2:])
#        print "(%d %d %0.2f)" % (n, sp, sp / (n / 2) ** 0.5),
#    print
#    print 'dec', Foo(dec, BLOCK)
#    print 'inc', Foo(dec + inc, BLOCK)
#    print 'black', Foo(dec + black, BLOCK)
#    print 'white', Foo(dec + white, BLOCK)
#    print 'newwhite', Foo(dec + newwhite, BLOCK)
#    dec = GX2(-5, 5) + GX2(1, 22)
#    incx = GX2(-5, 0)
#    oddx = GX2(-13, 0) + GX2(-10, 26)
#    evenx = GX2(-16, 0) + GX2(-15, 16)
#    dec = GX(-5, -15) + GX(1, -20)
#    incx = GX(-5, -10)
#    oddx = GX(-13, -26) + GX(-10, -46)
#    evenx = GX(-16, -32) + GX(-15, -46)
#    dec = GX(-5, 0) + GX(1, -5)
#    incx = GX(-5, 0)
#    oddx = GX(-13, 0) + GX(-10, -20)
#    evenx = GX(-16, 0) + GX(-15, -14)
#    dec = Salvo((G(-5, 0), G(1, 5)))
#    incx = G(-5, 0).deepenedTo(1)
#    oddx = Salvo((G(-13, 0), G(-10, 20)))
#    evenx = Salvo((G(-16, 0), G(-15, 14)))
#    #print incx, oddx, evenx
#    #Hoo(BLOCK, (dec, oddx, evenx))
#    #Hoo(BLOCK, (dec, incx, evenx))
#    Hoo(BLOCK, (dec, incx, oddx))
#    dec3 = GX(-6, 0) + GX(4, 0)
#    s1 = GX(-8, 0) + GX(-4, -11)
#    s2 = GX(-11, 0) + GX(-9, -24)
    #Hoo(BLOCK, (dec, s1, s2))
#    s3 = G(-9, 0).deepenedTo(1) # bingo
#    s4 = G(-4, 0).deepenedTo(1) # bingo
#    Hoo(BLOCK, (dec, s3, s4))
    #s5 = GX(-7, 0) # start from -12
    #Hoo(BLOCK, (dec, incx, s5))
    ############################################
    #g1 = GX(-5, 0)
    #g2 = GX(1, 0)
    #g3 = GX(-6, 0) # ok -3 -16
    #g3 = GX(-3, 0) # ok 7, possible -32 12
    #g3 = GX(4, 0) # possible -12, 1
    ############################################
    #g1 = GX(-5, 0)
    #g2 = GX(2, 0)
    #g3 = GX(-4, 0) # Done -10 to
    ############################################
#    g1 = G(-3, 0).deepenedTo(1)
#    g2 = G(-5, 0).deepenedTo(1)
    #g3 = GX(-2, 0) # -11, -2 but no 1c moves
#    g3 = G(2, 0).deepenedTo(1) # -11, 0 and -3,-2,-1,2 and 2c -5,1,3,8
    #g3 = GX(8, 0) # Done -10 to 8
    ############################################
#    g1 = GX(-5, 0)
#    g2 = GX(-4, 0)
#    g3 = GX(-3, 0) # Done -3 to
#    Hoo(BLOCK, (g1, g2, g3))

#    dec = Salvo(G(-5, 0), G(1, 5))
#    s3 = Salvo(G(-9, 0))
#    s4 = Salvo(G(-4, 0))
#    #print Searches[BLOCK, (dec, s3, s4)]
#    Assess(BLOCK, (dec, s3, s4), 1)

# TO DO
# Add shmear testing
# Eliminate redundant searches when two lanes are the same
# Add first-cycle non-elbow cleanups
# Add a third cycle search
# Add constructability results to results.py
# Deal better with slow gliders within cycles
# Write definitive outcome code

# DONE
# Allow one-time essential abort
# Prioritize rare search over common search for essentials
# Prioritize new bonus operations over those already seen
# Add farside inserters
# Fix parity error in p2 cleanup salvos
# Add second-cycle output detection
# Look for multiple cleanups of the same pattern
# Store cleanup search results in intermediate dictionary
# Reduce verbosity of cleanup reports
# Add forced-bonus search option even after no complete set found
